import requests


class Api:
    """
    Basic wrapper around the Github Api
    """

    def __init__(self):
        self._requests = requests

    def contents(self, user, repo, path=''):
        url = f'https://api.github.com/repos/{user}/{repo}/contents'
        if path:
            url += f'/{path}'

        response = self._get_url(url)
        return response.json()

    def download_raw_text_file(self, raw_text_file_url: str):
        response = self._get_url(raw_text_file_url)
        return response.text

    def _get_url(self, url: str):
        response = self._requests.get(url)
        response.raise_for_status()
        return response



from concurrent import futures
from pathlib import Path

from .api import Api
from ..models import DownloadableFile


class Repo:

    def __init__(self, api: Api, executor: futures.Executor):
        self._api = api
        self._executor = executor

    def file_urls(self, user, repo, path):
        """
        Explore recursively a repo and extract the file list

        :param user: Github Username
        :param repo: Github Repo
        :param path: Path in the Repo
        :return: Flat list of all files recursively found along with their download URLs
        """
        files = self._get_files_in_dir(user, repo, path)
        return self._format_result(files)

    def _get_files_in_dir(self, user, repo, dir_path):
        def filter_by_type(contents, content_type):
            return [entry for entry in contents if entry['type'] == content_type]

        def get_files_in_all_sub_dirs_async():
            sub_dir_files_futures = []
            for sub_dir in sub_dirs:
                sub_dir_path = f"{dir_path}/{sub_dir['name']}".lstrip('/')
                sub_dir_files_future = self._executor.submit(self._get_files_in_dir, user, repo, sub_dir_path)
                sub_dir_files_futures += [sub_dir_files_future]

            all_sub_dir_files = []
            for sub_dir_files_future in futures.as_completed(sub_dir_files_futures):
                sub_dir_files = sub_dir_files_future.result()
                all_sub_dir_files += sub_dir_files

            return all_sub_dir_files

        dir_contents = self._api.contents(user, repo, dir_path)
        files = filter_by_type(dir_contents, 'file')
        sub_dirs = filter_by_type(dir_contents, 'dir')
        return files + get_files_in_all_sub_dirs_async()

    @staticmethod
    def _format_result(contents):
        return [
            DownloadableFile(
                file_path=Path(file['path']),
                download_url=file['download_url']
            ) for file in contents]



from concurrent import futures
from concurrent.futures import Executor
from pathlib import Path
from typing import List, NamedTuple

from .github.api import Api
from .models import DownloadableFile


class _DownloadedFile(NamedTuple):
    file_path: Path
    file_text_contents: str


class Downloader:
    def __init__(self, api: Api, executor: Executor):
        self._api = api
        self._executor = executor

    def download_files_at_location(self, root_dir: Path, files_to_download: List[DownloadableFile]) -> None:
        if not root_dir.exists() or not root_dir.is_dir():
            raise ValueError(f"Root dir '{root_dir}' isn't a valid directory")

        download_file_futures = []
        for file_to_download in files_to_download:
            download_file_futures.append(
                self._executor.submit(self._download_file, file_to_download))

        for download_file_future in futures.as_completed(download_file_futures):
            downloaded_file = download_file_future.result()
            self._write_to_file_in_sub_path(root_dir, downloaded_file.file_path, downloaded_file.file_text_contents)

    def _download_file(self, file: DownloadableFile):
        file_contents = self._api.download_raw_text_file(file.download_url)
        return _DownloadedFile(file_path=file.file_path, file_text_contents=file_contents)

    @staticmethod
    def _write_to_file_in_sub_path(root_dir: Path, file_sub_path: Path, file_content: str):
        def create_dir_hierarchy_if_does_not_exist():
            file_full_path.parent.mkdir(parents=True, exist_ok=True)

        def write_to_file():
            with file_full_path.open('w') as file:
                file.write(file_content)

        file_full_path = root_dir / file_sub_path
        create_dir_hierarchy_if_does_not_exist()
        write_to_file()



from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from pprint import pprint
from typing import List

import click

from kata.io.downloader import Downloader
from kata.io.models import DownloadableFile
from .io.github.api import Api
from .io.github.repo import Repo


@click.command()
@click.argument('github_user')
@click.argument('repo')
@click.argument('sub_path_in_repo', default='')
def cli(github_user, repo, sub_path_in_repo):
    main = Main()
    # main.explore_github_repo(user=github_user, repo_name=repo, sub_path_in_repo=sub_path_in_repo)
    main.download_github_repo(user=github_user, repo_name=repo, sub_path_in_repo=sub_path_in_repo)


class Main:
    def __init__(self):
        self._executor = ThreadPoolExecutor(100)
        self._api = Api()
        self._repo_explorer = Repo(self._api, self._executor)
        self._downloader = Downloader(self._api, self._executor)

    def explore_github_repo(self, user, repo_name, sub_path_in_repo):
        click.echo('Debug - Print all files in repo')
        click.echo('')
        click.echo('Exploring:')
        click.echo(f" - User: '{user}'")
        click.echo(f" - Repo: '{repo_name}'")
        click.echo(f" - SubPath in Repo: '{sub_path_in_repo}'")
        click.echo('')
        result = self._repo_explorer.file_urls(user, repo_name, sub_path_in_repo)
        pprint(result)
        click.echo('')
        click.echo('Done')

    def download_github_repo(self, user, repo_name, sub_path_in_repo):
        sandbox = Path('./sandbox')
        sandbox.mkdir(exist_ok=True)
        click.echo(f'Sandbox: {sandbox.absolute()}')

        repo_files: List[DownloadableFile] = self._repo_explorer.file_urls(user, repo_name, sub_path_in_repo)
        click.echo('Finished fetching the list. Writing to drive now')
        self._downloader.download_files_at_location(sandbox, repo_files)
        click.echo('Done! (probably ^_^)')



from setuptools import find_packages, setup


def readme():
    with open('README.md') as f:
        return f.read()


setup(
    name='kata',
    version='0.0.7',
    description='Kata made easy: A TDD setup in the language of your choice in a single command',
    long_description=readme(),
    keywords='test tdd kata clean-code softwarecraft',
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Topic :: Utilities',
        'Topic :: Software Development :: Testing'
    ],
    url='https://floriankempenich.github.io/kata',
    author='Florian Kempenich',
    author_email='Flori@nKempenich.com',
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    license='MIT',
    scripts=['bin/kata'],
    install_requires=[
        'click',
        'requests'
    ],
    include_package_data=True
)
